#version 430 core

// 1/pi
#define PI_RCP 0.31831

in vec2 FragTextureCoords;

out vec4 FragColor;

uniform sampler2D shadowMap;
uniform sampler2D image;
uniform sampler2D depthMap;
uniform sampler2D refractionFrontNormal;
uniform sampler2D refractionFrontDepth;
uniform sampler2D refractionFrontColor;
uniform sampler2D refractionBackNormal;
uniform sampler2D refractionBackDepth;
uniform sampler2D refractionBackColor;
uniform sampler2D rrColor;
uniform sampler2D rrDepth;
uniform sampler2D reflectionRefraction;

uniform mat4 lightSpaceMatrix;
uniform mat4 lightSpaceViewMatrix;
uniform mat4 lightSpaceProjMatrix;
uniform vec3 lightPosition;
uniform vec3 lightColor = vec3(0.0);
uniform mat4 viewMatrix;
uniform mat4 projectionMatrix;
uniform mat4 inverseViewMatrix;
uniform mat4 inverseProjectionMatrix;
uniform float Scale; // Scale of models for length dependent operations

const float step_size = 1.0;
const int max_samples = 40;

float shadowTest(vec4 lightSpaceCoords)
{
	vec4 projectedLightSpaceCoords = lightSpaceProjMatrix * lightSpaceCoords;
	vec3 projectedCoords = (projectedLightSpaceCoords.xyz / projectedLightSpaceCoords.w) * 0.5 + 0.5;
	
	// Samples outside of spotlight outerRadius cone defined as in shadow
	vec3 plsc = projectedLightSpaceCoords.xyz / projectedLightSpaceCoords.w;
	if (plsc.x * plsc.x + plsc.y * plsc.y > 1.0)
		return 1.0;
	
	// ShadowMap Lookup
	float referenceDepth = texture(shadowMap, projectedCoords.xy).r;

	if (projectedCoords.z > referenceDepth) 
		return 1.0;
	return 0.0;
}

vec3 getLightSpaceCoords(vec4 worldSpaceCoordsH)
{
	vec4 lightSpaceCoordsH = lightSpaceViewMatrix * worldSpaceCoordsH;
	return lightSpaceCoordsH.xyz / lightSpaceCoordsH.w;
}

vec4 getCameraSpaceCoords(vec3 screenSpaceCoords)
{
	return inverseProjectionMatrix * vec4(screenSpaceCoords * 2.0 - 1.0, 1.0);
}

vec4 getWorldSpaceCoords(vec3 screenSpaceCoords)
{
	return inverseViewMatrix * getCameraSpaceCoords(screenSpaceCoords);
}

vec4 sampleVolumetricLight()
{
	// Test values
	float phi = 1.0f;
	float tau = 1.0f; // TODO Check other possible values
	float raymarch_distance_worldspace = 20.0;
	float step_size_worldspace = 0.1;
	float light_contribution = 0.0f;

	vec3 endPointInScreenSpaceA = vec3(FragTextureCoords, 0.0f);
	vec3 endPointInScreenSpaceB = vec3(FragTextureCoords, texture(depthMap, FragTextureCoords).r);
	vec3 endPointInLightSpaceA = getLightSpaceCoords(getWorldSpaceCoords(endPointInScreenSpaceA));
	vec3 endPointInLightSpaceB = getLightSpaceCoords(getWorldSpaceCoords(endPointInScreenSpaceB));
	vec3 ray_direction = normalize(endPointInLightSpaceB - endPointInLightSpaceA);
	//vec3 lightPositionLS = getLightSpaceCoords(vec4(lightPosition, 1.0));
	
	// Shorten raymarch distance to closest intersection (no need to sample behind polygons)
	vec4 fragPosInCameraSpace = getCameraSpaceCoords(endPointInScreenSpaceB);
	float fragDepthInCameraSpace = -fragPosInCameraSpace.z / fragPosInCameraSpace.w;
	if (fragDepthInCameraSpace < raymarch_distance_worldspace) raymarch_distance_worldspace = fragDepthInCameraSpace;

	vec3 samplePosition = endPointInLightSpaceA;
	
	for (float l = raymarch_distance_worldspace - step_size_worldspace; l > step_size_worldspace; l -= step_size_worldspace)
	{
		// Light position in Light Space is (0, 0, 0)
		float d = length(samplePosition) * Scale;
		float d_rcp = 1.0/d;

		light_contribution += tau * ((1.0 - shadowTest(vec4(samplePosition, 1.0))) * (phi * 0.25 * PI_RCP) * d_rcp * d_rcp ) * exp(-d*tau) * exp(-(l*Scale)*tau) * step_size_worldspace;

		samplePosition += step_size_worldspace * ray_direction;
	}

	return vec4(lightColor * light_contribution, 0.0);
}

vec3 wToSS(vec3 worldPos)
{
    vec4 ssPos = projectionMatrix * viewMatrix * vec4(worldPos, 1.0f);
	vec3 ss =  (ssPos.xyz /= ssPos.w)*0.5+0.5;
    return ss;
}

vec4 raymarchSSR(vec3 FragCoordsWS, vec3 raymarch_direction, float step_size, int max_raymarch_samples)
{
	// raymarch
	vec3 raySample = FragCoordsWS;
	vec2 reflectSample = FragTextureCoords;
	
	for (int i = 0; i <= max_raymarch_samples; i++)
	{
		//raySample += step_size * raymarch_direction;
		raySample = FragCoordsWS + (exp(0.02*i)-1.0)*raymarch_direction;
		vec4 SampleCoordsVS = projectionMatrix * viewMatrix * vec4(raySample, 1);
		vec3 SampleCoordsSS = (SampleCoordsVS.xyz / SampleCoordsVS.w)*0.5 + 0.5;
		
		// Check if ray leaves image texture area
		if (SampleCoordsSS.x < 0.0 || SampleCoordsSS.x > 1.0 || SampleCoordsSS.y < 0.0 || SampleCoordsSS.y > 1.0) return vec4(0.0);
		
		// Check if ray intersects object (break and use last sample)
		if (SampleCoordsSS.z > texture(depthMap, SampleCoordsSS.xy).r) 
			break;
			
		reflectSample = SampleCoordsSS.xy;
		//if (i == max_raymarch_samples) return vec4(0.0);
	}
	
	// hide edges
	float strength = 1.0f;
	if (reflectSample.y > 0.5) strength *= (2.0 - reflectSample.y*2.0);
	if (reflectSample.y < 0.5) strength *= (reflectSample.y*2.0);
	//if (reflectSample.x > 0.9) strength *= (10.0 - reflectSample.x*10.0);
	//if (reflectSample.x < 0.1) strength *= (reflectSample.x*10.0);

	return vec4(texture(image, reflectSample).xyz,0)*strength;
}

vec4 sampleSSR(vec3 SurfaceCoordsWS, vec3 eye_ray_direction)
{
	vec4 baseWaterColor = vec4(0.1,0.1,0.2,0);
	//vec4 ScreenCoordsWSH = getWorldSpaceCoords(vec3(FragTextureCoords, 0.0));
	//vec3 ScreenCoordsWS = ScreenCoordsWSH.xyz / ScreenCoordsWSH.w;
	
	vec3 surfaceNormal = vec3(0.0,1.0,0.0); // normalized
	//vec3 eyeRayDir = normalize(FragCoordsWS - ScreenCoordsWS);
	

	// Reflection
	// Reflect eyeRay in frag point on frag surface normal
	vec3 reflectRayDir = eye_ray_direction - 2.0 * dot(eye_ray_direction,surfaceNormal) * surfaceNormal;
	vec4 reflectionCol = vec4(0.0);
	if (dot(eye_ray_direction,reflectRayDir) > 0) // Check if reflected area is behind camera
		reflectionCol = raymarchSSR(SurfaceCoordsWS, reflectRayDir, 1.0, 200);


	// Refraction
	vec3 refractionRayDir = refract(eye_ray_direction, surfaceNormal, 0.752); // refractive index of air=1.0 / water=1.33
	vec4 refractionCol = raymarchSSR(SurfaceCoordsWS, refractionRayDir, 0.05, 200);
	

	return 0.5*baseWaterColor + 0.2*reflectionCol + refractionCol;
}

vec3 traceRay(vec3 FragCoordsWS, vec3 raymarch_direction) 
{
    /*vec3 hit_color = vec3(0.0f);
    vec3 step = raymarch_direction * step_size;
    vec3 sample_pos = FragCoordsWS;

    for (int i = 0; i < 100; i++) {
        sample_pos += step;

        vec3 screen_space = wToSS(sample_pos);
		
		float sample_depth = getWorldSpaceCoords(vec3(screen_space.xy,texture(depthMap,screen_space.xy).r)).z;
        if (sample_pos.z > sample_depth && sample_pos.z - sample_depth < 2.0f) {
            for (int k = 0; k < 6; k++) {
                step *= 0.25f;
				screen_space = wToSS(sample_pos);

				sample_depth = getWorldSpaceCoords(vec3(screen_space.xy,texture(depthMap,screen_space.xy).r)).z;
                if (sample_pos.z < sample_depth)
				{
                    sample_pos += step;
                } else 
				{
                    sample_pos -= step;
                }
            }
            if (screen_space.x >= 0.0f && screen_space.x <= 1.0f && screen_space.y >= 0.0f && screen_space.y <= 1.0f)
            {
                hit_color = texture(image, screen_space.xy).xyz;
            }
            break;
        }
		if (sample_pos.z > sample_depth) 
			break;
		hit_color = texture(image, screen_space.xy).xyz;
    }
    return hit_color;*/

	vec3 raySample = FragCoordsWS;
	vec2 reflectSample = FragTextureCoords;
	vec3 step = step_size * raymarch_direction;
	for (int i = 0; i <= max_samples; i++)
	{
		//raySample += step_size * raymarch_direction;
		raySample += step;
		vec4 SampleCoordsVS = projectionMatrix * viewMatrix * vec4(raySample, 1);
		vec3 SampleCoordsSS = (SampleCoordsVS.xyz / SampleCoordsVS.w)*0.5 + 0.5;
		
		if (SampleCoordsSS.x < 0.0 || SampleCoordsSS.x > 1.0 || SampleCoordsSS.y < 0.0 || SampleCoordsSS.y > 1.0) return vec3(0.0);
		
		float sample_depth = texture(depthMap, SampleCoordsSS.xy).r;
		
		if (SampleCoordsSS.z > sample_depth && SampleCoordsSS.z - sample_depth < 2.0) {
            for (int k = 0; k < 6; k++) {
                step *= 0.5;
				SampleCoordsVS = projectionMatrix * viewMatrix * vec4(raySample, 1);
				SampleCoordsSS = (SampleCoordsVS.xyz / SampleCoordsVS.w)*0.5 + 0.5;

				sample_depth = texture(depthMap, SampleCoordsSS.xy).r;
                if (SampleCoordsSS.z < sample_depth)
				{
                    raySample += step;
                } else 
				{
                    raySample -= step;
                }
            }
            if (SampleCoordsSS.x >= 0.0f && SampleCoordsSS.x <= 1.0f && SampleCoordsSS.y >= 0.0f && SampleCoordsSS.y <= 1.0f)
            {
                reflectSample = SampleCoordsSS.xy;
            }
            break;
        }
			
		reflectSample = SampleCoordsSS.xy;
		//if (i == max_raymarch_samples) return vec4(0.0);
	}
	
	// hide edges
	float strength = 1.0f;
	if (reflectSample.y > 0.5) strength *= (2.0 - reflectSample.y*2.0);
	if (reflectSample.y < 0.5) strength *= (reflectSample.y*2.0);
	//if (reflectSample.x > 0.9) strength *= (10.0 - reflectSample.x*10.0);
	//if (reflectSample.x < 0.1) strength *= (reflectSample.x*10.0);

	return texture(image, reflectSample).xyz*strength;
}

vec2 traceBackfaceRay(vec3 FragCoordsWS, vec3 raymarch_direction) 
{
	vec3 raySample = FragCoordsWS;
	vec2 reflectSample = FragTextureCoords;
	vec3 step = step_size/10 * raymarch_direction;
	for (int i = 0; i <= 20; i++)
	{
		//raySample += step_size * raymarch_direction;
		raySample += step;
		vec4 SampleCoordsVS = projectionMatrix * viewMatrix * vec4(raySample, 1);
		vec3 SampleCoordsSS = (SampleCoordsVS.xyz / SampleCoordsVS.w)*0.5 + 0.5;
		
		if (SampleCoordsSS.x < 0.0 || SampleCoordsSS.x > 1.0 || SampleCoordsSS.y < 0.0 || SampleCoordsSS.y > 1.0) return vec2(0.0);
		
		float sample_depth = texture(refractionBackDepth, SampleCoordsSS.xy).r;
		
		if (SampleCoordsSS.z > sample_depth && SampleCoordsSS.z - sample_depth < 2.0) {
            for (int k = 0; k < 12; k++) {
                step *= 0.5;
				SampleCoordsVS = projectionMatrix * viewMatrix * vec4(raySample, 1);
				SampleCoordsSS = (SampleCoordsVS.xyz / SampleCoordsVS.w)*0.5 + 0.5;

				sample_depth = texture(refractionBackDepth, SampleCoordsSS.xy).r;
                if (SampleCoordsSS.z < sample_depth)
				{
                    raySample += step;
                } else 
				{
                    raySample -= step;
                }
            }
            if (SampleCoordsSS.x >= 0.0f && SampleCoordsSS.x <= 1.0f && SampleCoordsSS.y >= 0.0f && SampleCoordsSS.y <= 1.0f)
            {
                reflectSample = SampleCoordsSS.xy;
            }
            break;
        }
			
		reflectSample = SampleCoordsSS.xy;
		//if (i == max_raymarch_samples) return vec4(0.0);
	}

	float strength = 1.0f;
	if (reflectSample.y > 0.5) strength *= (2.0 - reflectSample.y*2.0);
	if (reflectSample.y < 0.5) strength *= (reflectSample.y*2.0);


	return reflectSample;
}

void main()
{
	float refraction = texture(reflectionRefraction, FragTextureCoords).r;
	float reflection = texture(reflectionRefraction, FragTextureCoords).g;
	float transparency = texture(reflectionRefraction, FragTextureCoords).b;
	float depth = texture(depthMap, FragTextureCoords).r;
	float rrDepth = texture(rrDepth, FragTextureCoords).r;
	if((refraction == 0.0 && reflection == 0.0) || depth < rrDepth)
	{	
		FragColor = texture(image, FragTextureCoords) + sampleVolumetricLight();;
		return;
	}
	else
	{	
		vec4 ScreenCoordsWSH = getWorldSpaceCoords(vec3(FragTextureCoords, 0.0));
		vec3 ScreenCoordsWS = ScreenCoordsWSH.xyz / ScreenCoordsWSH.w;

		vec4 FragCoordsWSH = getWorldSpaceCoords(vec3(FragTextureCoords, rrDepth));
		vec3 FragCoordsWS = FragCoordsWSH.xyz / FragCoordsWSH.w;
		
		vec3 eyeRayDir = normalize(FragCoordsWS - ScreenCoordsWS);

		vec3 normal = (2.0*texture(refractionFrontNormal, FragTextureCoords).xyz)-1.0;
		vec3 reflection_color;
		vec3 refraction_color;
		if(refraction > 0.0)
		{
			float back_depth = texture(refractionBackDepth, FragTextureCoords).r;
			if(back_depth < 1.0)
			{
				float front_depth = texture(refractionFrontDepth, FragTextureCoords).r;
				float d = front_depth - back_depth;
				float scaled_d = 0.1 + (200.0 - 0.1) * d; 
				vec3 refraction_dir = refract(normalize(eyeRayDir),normalize(normal),1/1.2);
				//vec3 p2 = FragCoordsWS + scaled_d * refraction_dir;
				//vec3 p2_ss = wToSS(p2);
				vec2 back_tex = traceBackfaceRay(FragCoordsWS, refraction_dir);

				vec4 backFragCoordsWSH = getWorldSpaceCoords(vec3(back_tex, texture(refractionBackDepth,back_tex)));
				vec3 backFragCoordsWS = backFragCoordsWSH.xyz / backFragCoordsWSH.w;

				vec3 back_normal = (2.0*texture(refractionBackNormal, back_tex).xyz)-1.0;
				vec3 back_refraction_dir = refract(normalize(refraction_dir),normalize(-back_normal),1.2/1);
				refraction_color = traceRay(backFragCoordsWS, back_refraction_dir);
			}
			else
			{
				vec3 refraction_dir = refract(eyeRayDir,normal,1/1.33);
				refraction_color = traceRay(FragCoordsWS, refraction_dir);
			}
		}	
		if(reflection > 0.0)
		{
			vec3 reflection_dir = normalize(reflect(eyeRayDir, normal));
			if (dot(eyeRayDir,reflection_dir) > 0)
				reflection_color = traceRay(FragCoordsWS, reflection_dir);
		}
		FragColor = (1-transparency)*(reflection*vec4(reflection_color,1.0) + (1-reflection) * texture(rrColor,FragTextureCoords)) + transparency * vec4(refraction_color,1.0);
	}
	
	/*// ScreenCoords represent the point where the eyeRay starts
	vec4 ScreenCoordsWSH = getWorldSpaceCoords(vec3(FragTextureCoords, 0.0));
	vec3 ScreenCoordsWS = ScreenCoordsWSH.xyz / ScreenCoordsWSH.w;
	// FragCoords represent the point where the eyeRay intersects the first solid object
	vec4 FragCoordsWSH = getWorldSpaceCoords(vec3(FragTextureCoords, texture(depthMap, FragTextureCoords).r));
	vec3 FragCoordsWS = FragCoordsWSH.xyz / FragCoordsWSH.w;
	
	vec3 eyeRayDir = normalize(FragCoordsWS - ScreenCoordsWS);
	
	
	// Only lake is reflective
	if (texture(depthMap, FragTextureCoords).r < 0.9999 && FragCoordsWS.y < -4.4)
	{
		vec3 vnormEyeRayDir = -eyeRayDir / eyeRayDir.y;
		vec3 SurfaceCoordsWS = ScreenCoordsWS + (ScreenCoordsWS.y + 4.4)*vnormEyeRayDir; // Calculate Point on water surface
		FragColor = sampleSSR(SurfaceCoordsWS, eyeRayDir);
	}
	else
		FragColor = texture(image, FragTextureCoords) + sampleVolumetricLight();*/
}