#version 430 core
in vec2 vertexTexCoord;
in vec4 FragPosition;
in vec4 FragPositionLightSpace;
in vec3 Normal;

layout(location = 0) out vec4 FragColor;

#define MAX_NUM_POINT_LIGHTS 100
#define MAX_NUM_SPOT_LIGHTS 100

struct PointLight 
{
    vec3 position;
    vec3 color;
};

struct SpotLight 
{
    vec3 position;
    vec3 direction;
    vec3 color;
    float innerRadius;
    float outerRadius;     
};

struct DirectionLight
{
    vec3 position;
    vec3 direction;
    vec3 color;
};

uniform sampler2D ourTexture;
uniform sampler2D shadowMap;
uniform int numPointLights;
uniform int numSpotLights;
uniform PointLight pointLights[MAX_NUM_POINT_LIGHTS];
uniform SpotLight spotLights[MAX_NUM_SPOT_LIGHTS];
uniform DirectionLight dirLight;
uniform float ambientStrength;
uniform vec3 ambientColor;

float shadowTest(vec3 projectedCoords, vec2 offset)
{
	float bias = 0.001;
	
	// No shadow on fragments outside of shadowmap (for now)
	if (projectedCoords.x < 0 || projectedCoords.x > 1 || projectedCoords.y < 0 || projectedCoords.y > 1)
		return 0.0;
	
	float referenceDepth = texture(shadowMap, projectedCoords.xy + offset).r; // depth value stored in first channel
	if (projectedCoords.z - bias > referenceDepth) 
		return 1.0;
	return 0.0;
}

vec3 getProjLightSpaceCoords(vec4 fragPositionLightSpace)
{
	vec3 projectedCoords = fragPositionLightSpace.xyz / fragPositionLightSpace.w;
	return projectedCoords * 0.5 + 0.5; // transform to NDC
}

float shadowTestPCF(vec3 projectedCoords)
{
	float shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
	for (int i = -1; i < 2; i++)
	{
		for (int j = -1; j < 2; j++)
		{
			shadow += shadowTest(projectedCoords, vec2(i, j) * texelSize);
		}
	}
	return shadow / 9.0;
}

void main()
{
	vec3 norm = normalize(Normal);
    vec3 diff = vec3(0,0,0);
    for(int i = 0; i < numPointLights; i++)
    {
        PointLight p = pointLights[i];
        vec3 lightDir = normalize(p.position - FragPosition.xyz);
        diff += max(dot(norm, lightDir), 0.0) * p.color;
    }
    for(int i = 0; i < numSpotLights; i++)
    {
        SpotLight s = spotLights[i];
        vec3 lightDir = normalize(s.position - FragPosition.xyz);
        float theta = dot(lightDir, normalize(-s.direction));
		float shadow = shadowTestPCF(getProjLightSpaceCoords(FragPositionLightSpace));
        if(theta > s.innerRadius)
        {       
            diff += max(dot(norm, lightDir), 0.0) * s.color * (1.0 - shadow);
        }
		else if (theta > s.outerRadius)
		{
			diff += max(dot(norm, lightDir), 0.0) * (theta - s.outerRadius)/(s.innerRadius - s.outerRadius) * s.color * (1.0 - shadow);
		}
    }
	//float shadow = shadowTestPCF(getProjLightSpaceCoords(FragPositionLightSpace));

	FragColor = texture(ourTexture, vertexTexCoord) * (vec4(diff,0.0f) + (ambientStrength * vec4(ambientColor,1.0f)));
}
